"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parallelLimit = void 0;
const queue_1 = require("@zodash/queue");
const event_1 = require("@zodash/event");
const cache_1 = require("@zodash/cache");
const generatee_1 = require("@zodash/generatee");
const next_tick_1 = require("@zodash/next-tick");
// export interface Queue<T> {
//   enqueue(value: T): void;
//   dequeue(): T;
//   size(): number;
//   isEmpty(): boolean;
// }
// const nextTick = (callback: (...args: any) => void, ...args: any) => typeof setImmediate !== 'undefined' ? setImmediate(callback, ...args) : setTimeout(callback, 0, ...args);
// const poll = (queue: Queue<ITask<any>>) => {
//   if (queue.isEmpty()) {
//     nextTick(poll, queue);
//   }
//   const task = queue.dequeue();
//   task((err, result) => {
//     nextTick(poll, queue);
//   });
// };
// const poll = (generators: any) => {
//   if (generators.done()) {
//     nextTick(poll, generators);
//   }
//   const task = generators.next();
//   task((err, result) => {
//     nextTick(poll, generators);
//   });
// };
function toPromise(fn) {
    if (fn.length === 1) {
        return () => {
            return new Promise((resolve, reject) => {
                return fn.call(null, (error, result) => {
                    if (error) {
                        return reject(error);
                    }
                    return resolve(result);
                });
            });
        };
    }
    return fn;
}
function parallelLimit(tasks, limit, cb) {
    const store = new cache_1.Cache(Infinity);
    const emitter = new event_1.Event();
    const running = new queue_1.Queue(limit);
    const results = [];
    const taskIds = [];
    tasks.forEach((task, index) => {
        const id = index;
        taskIds.push(id);
        store.set(id, { id, index, task });
    });
    const source = generatee_1.generatee(taskIds);
    function done() {
        if (cb) {
            cb(results);
            cb = null;
        }
        else {
            emitter.emit('done', results);
        }
    }
    function next() {
        if (running.isFull()) {
            return next_tick_1.nextTick(next);
        }
        const data = source.next();
        // no rest tasks
        if (data.done) {
            // no running tasks
            if (!running.isEmpty()) {
                return;
            }
            return next_tick_1.nextTick(done);
        }
        const value = data.value;
        running.enqueue(value);
        const taskId = value;
        const { task, index } = store.get(taskId);
        toPromise(task)()
            .then(function _done(result) {
            results[index] = result;
        })
            .catch(function _done(error) {
            results[index] = error;
        })
            .finally(() => {
            running.dequeue();
            // give err and result to next
            next_tick_1.nextTick(next);
        });
    }
    function setup() {
        if (!tasks.length) {
            // empty
            return done();
        }
        return next_tick_1.nextTick(next);
    }
    // setup
    setup();
    if (!cb) {
        return new Promise((resolve) => {
            emitter.on('done', resolve);
        });
    }
}
exports.parallelLimit = parallelLimit;
//# sourceMappingURL=parallel-limit.js.map